/* $Id$ */
/*
 * Copyright (c) 2003-2004 Jose Nazario <jose@monkey.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Jose Nazario.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <event.h>
#include <dnet.h>
#include <pcap.h>

#include "jtrace.h"
#include "pcaputil.h"

void
usage(char *progname)
{
	printf("usage: %s -d dest [-f hop] [-p port] [-i intf]\n", progname);
	exit(0);
}

/* stolen from openbsd libc, needed my own type */

char           *
inet_ntoasc(ip_addr_t in)
{
	static char     b[18];
	register char  *p;

	p = (char *) &in;
#define UC(b)        (((int)b)&0xff)
	(void) snprintf(b, sizeof(b),
		     "%u.%u.%u.%u", UC(p[0]), UC(p[1]), UC(p[2]), UC(p[3]));
	return (b);
}

static void
report(u_char * packet, void *arg)
{
	struct myctx   *ctx = (struct myctx *) arg;
	struct ip_hdr  *ip_h;
	struct tcp_hdr *tcp;
	struct icmp_hdr *icmp;
	u_char         *tmp;
	const char     *p;
	char 	       *reply = "U!";
	struct addr     ip_src;

	tmp = packet + ctx->dl_len;
	ip_h = (struct ip_hdr *) tmp;
	if (ip_h->ip_v != 4)
		return;
	p = inet_ntoasc(ip_h->ip_src);
	if ((addr_aton(p, &ip_src)) == -1)
		return;
	switch(ip_h->ip_p) {
	case IP_PROTO_TCP:
		if (addr_cmp(&ip_src, &(ctx->dst)) != 0) 
			return;

		tcp = (struct tcp_hdr *)(tmp + IP_HDR_LEN);

		if (((tcp->th_flags == TH_SYN + TH_ACK) || 
		     (tcp->th_flags == TH_RST + TH_ACK)) &&
		     (htons(tcp->th_sport) == ctx->dport))  {
			char *flags = "SA";

			if (tcp->th_flags == TH_RST + TH_ACK)
				flags = "R";
			printf("%3d: %-16s %-5s\n", ctx->hop, 
				addr_ntoa(&ip_src), flags); 
			event_del(&ctx->recv_ev);
		}
		break;
	case IP_PROTO_ICMP:
		icmp = (struct icmp_hdr *)(tmp + IP_HDR_LEN);
		/* look for TTL_EXPIRED, PROHIBITED msgs */

		if (icmp->icmp_type == ICMP_UNREACH) {
			switch (icmp->icmp_code) {
			case ICMP_UNREACH_NET:
				reply = "N!";
				event_del(&ctx->recv_ev);
				break;
			case ICMP_UNREACH_HOST:
				reply = "H!";
				event_del(&ctx->recv_ev);
				break;
			case ICMP_UNREACH_PROTO:
				reply = "P!";
				event_del(&ctx->recv_ev);
				break;
			case ICMP_UNREACH_NET_PROHIB:
			case ICMP_UNREACH_HOST_PROHIB:
				reply = "A!";
				event_del(&ctx->recv_ev);
				break;
			}
		} else if (icmp->icmp_type == ICMP_TIMEXCEED) {
			reply = "X!";
			event_add(&ctx->send_ev, &ctx->tv);
		}

		printf("%3d: %-16s %-5s\n", ctx->hop, addr_ntoa(&ip_src), 
			reply); 
		ctx->hop++;
		break;
	}
	fflush(stdout); 
}

static void
_recv(int fd, short event, void *arg)
{
	struct myctx   *ctx = (struct myctx *) arg;
	struct pcap_pkthdr ph;
	u_char         *pread;

	/* reschedule */
	event_add(&ctx->recv_ev, &ctx->tv);	

	if ((pread = (u_char *) pcap_next(ctx->p, &ph)) != NULL)
		report(pread, ctx);

	return;
}

static void
_send(int fd, short event, void *arg)
{
	struct myctx   *ctx = (struct myctx *)arg;
	struct jtrace_pkt *pkt;
	u_char          buf[BUFSIZ];
	int             len = IP_HDR_LEN + TCP_HDR_LEN;

	pkt = (struct jtrace_pkt *)buf;

	ip_pack_hdr(&pkt->pkt_hdr_i.ip, IP_TOS_LOWDELAY, len, 
		    rand_uint16(ctx->rand), 0, ctx->hop, IP_PROTO_TCP, 
		    ctx->src.addr_ip, ctx->dst.addr_ip);
	tcp_pack_hdr(&pkt->pkt_hdr_t.tcp, rand_uint16(ctx->rand), ctx->dport, 
		     rand_uint32(ctx->rand), rand_uint32(ctx->rand), 
		     TH_SYN, rand_uint16(ctx->rand), 0);

	ip_checksum(pkt, len);
	if (ip_send(ctx->ip, pkt, len) < 0)
		warn("ip_send in _send, short send");
	ctx->count++;
	return;
}

int
main(int argc, char *argv[])
{
	struct myctx    ctx;
	char            ebuff[PCAP_ERRBUF_SIZE];
	int             p_fd, ch;
	char           *intf = "fxp0", *dst = "127.0.0.1", buf[1024];
	time_t          start;
	intf_t 	       *my_intf;
	struct intf_entry *entry;

	ctx.hop = 1;
	ctx.dport = 80;

	if (argc < 2)
		usage(argv[0]);

	while ((ch = getopt(argc, argv, "d:f:i:p:t:h")) != -1) {
		switch (ch) {
		case 'd':
			dst = optarg;
			if (addr_aton(dst, &ctx.dst) == -1)
				err(1, "malformed destination address.");
			break;
		case 'f':
			ctx.hop = atoi(optarg);
			break;
		case 'p':
			ctx.dport = atoi(optarg);
			break;
		case 'i':
			intf = optarg;
			break;
		default:
			usage(argv[0]);
			break;
		}
	}
	argc -= optind;
	argv += optind;

	ctx.p = pcap_open_live(intf, 1500, 1, 500, ebuff);
	if (ctx.p == NULL)
		err(1, "pcap_open_live()");

	ctx.dl_len = pcap_dloff(ctx.p);

	event_init();

	ctx.tv.tv_sec = 0;
	ctx.tv.tv_usec = 500;

	entry = (struct intf_entry *)buf;
	memset(entry, 0, sizeof(*entry));
	entry->intf_len = sizeof(buf);
	my_intf = intf_open();
	if (intf_get_dst(my_intf, entry, &ctx.dst) < 0)
		err(1, "intf_get_src()");
	ctx.src = entry->intf_addr;
	intf_close(my_intf);

	p_fd = pcap_fileno(ctx.p);
	event_set(&ctx.recv_ev, p_fd, EV_READ, _recv, (void *)&ctx);
	event_add(&ctx.recv_ev, &ctx.tv);
	ctx.ip = ip_open();
	if (ctx.ip == NULL)
		err(1, "ip_open() failed ..");
	event_set(&ctx.send_ev, p_fd, EV_WRITE, _send, (void *)&ctx);
	event_add(&ctx.send_ev, &ctx.tv);

	ctx.rand = rand_open();

	start = time(NULL);

	printf("tracing from %s to %s on TCP port %d\n",
		addr_ntoa(&ctx.src), addr_ntoa(&ctx.dst), ctx.dport);
	fflush(stdout);
	ctx.count = 0;
	event_dispatch();

	printf("trace completed. total execution time was %d seconds.\n",
	       time(NULL) - start);
	ip_close(ctx.ip);
	rand_close(ctx.rand);
	return (1);
}
